import logging

import numpy as np

import aspire.image
from aspire.nufft import anufft
from aspire.numeric import fft, xp

logger = logging.getLogger(__name__)


class Sinogram:
    def __init__(self, data, dtype=None):
        """
        Initialize a Sinogram Object. This is a stack of one or more line projections or sinograms.

        The stack can be multidimensional with 'self.n' equal to the product
        of the stack dimensions. Singletons will be expanded into a stack
        with one entry.

        :param data: Numpy array containing image data with shape
            `(..., angles, radial points)`.
        :param dtype: Optionally cast `data` to this dtype.
            Defaults to `data.dtype`.

        :return: Sinogram instance holding `data`.
        """
        if dtype is None:
            self.dtype = data.dtype
        else:
            self.dtype = np.dtype(dtype)

        if data.ndim == 2:
            data = data[np.newaxis, :, :]
        if data.ndim < 3:
            raise ValueError(
                f"Invalid data shape: {data.shape}. Expected shape: (..., angles, radial_points), where '...' is the stack number."
            )

        self._data = data.astype(self.dtype, copy=False)
        self.ndim = self._data.ndim
        self.shape = self._data.shape
        self.stack_shape = self._data.shape[:-2]
        self.stack_n_dim = self._data.ndim - 2
        self.n = np.prod(self.stack_shape)
        self.n_angles = self._data.shape[-2]
        self.n_radial_points = self._data.shape[-1]

        # Numpy interop
        # https://numpy.org/devdocs/user/basics.interoperability.html#the-array-interface-protocol
        self.__array_interface__ = self._data.__array_interface__
        self.__array__ = self._data

    def _check_key_dims(self, key):
        if isinstance(key, tuple) and (len(key) > self._data.ndim):
            raise ValueError(
                f"Sinogram stack_dim is {self.stack_n_dim}, slice length must be =< {self.n_dim}"
            )

    def __getitem__(self, key):
        self._check_key_dims(key)
        return self.__class__(self._data[key])

    def __setitem__(self, key, value):
        self._check_key_dims(key)
        self._data[key] = value

    def stack_reshape(self, *args):
        """
        Reshape the stack axis.

        :*args: Integer(s) or tuple describing the intended shape.

        :return: Sinogram instance
        """

        # If we're passed a tuple, use that
        if len(args) == 1 and isinstance(args[0], tuple):
            shape = args[0]
        else:
            # Otherwise use the variadic args
            shape = args

        # Sanity check the size
        if shape != (-1,) and np.prod(shape) != self.n:
            raise ValueError(
                f"Number of sinogram images {self.n} cannot be reshaped to {shape}."
            )

        return self.__class__(self._data.reshape(*shape, *self._data.shape[-2:]))

    def asnumpy(self):
        """
        Return image data as a (<stack>, angles, radians)
        read-only array view.

        :return: read-only ndarray view
        """

        view = self._data.view()
        view.flags.writeable = False
        return view

    def copy(self):
        return self.__class__(self._data.copy())

    def __str__(self):
        return f"Sinogram(n_images = {self.n}, n_angles = {self.n_angles}, n_radial_points = {self.n_radial_points})"

    def __repr__(self):
        msg = f"Sinogram: {self.n} images of dtype {self.dtype}, "
        msg += f"arranged as a stack with shape {self.stack_shape}. "
        msg += f"Each image has {self.n_angles} angles and {self.n_radial_points} radial points."
        return msg

    def backproject(self, angles):
        """
        Backprojection method for a single stack of lines.

        :param angles: np.ndarray
            1D array of angles in radians. Each entry in the array
            corresponds to different angles which are used to
            reconstruct the image.
        :return: An Image object containing the original stack size
            with a newly reconstructed numpy array of the images.
            Expected return shape should be (..., n_radial_points, n_radial_points)
        """
        if len(angles) != self.n_angles:
            raise ValueError("Number of angles must match the number of projections.")

        original_stack_shape = self.stack_shape
        sinogram = xp.asarray(self.stack_reshape(-1)._data)
        L = self.n_radial_points
        sinogram = fft.ifftshift(sinogram, axes=-1)
        sinogram_ft = fft.rfft(sinogram, axis=-1)
        sinogram_ft *= xp.pi  # Fix scale to match
        sinogram_ft[..., 0] /= 2  # Fix DC
        angles = xp.asarray(angles)

        # grid generation with real points
        y_idx = fft.rfftfreq(self.n_radial_points) * xp.pi * 2
        n_real_points = len(y_idx)
        pts = xp.empty((2, len(angles), n_real_points), dtype=self.dtype)
        pts[0] = y_idx[xp.newaxis, :] * xp.sin(angles)[:, xp.newaxis]
        pts[1] = y_idx[xp.newaxis, :] * xp.cos(angles)[:, xp.newaxis]

        imgs = anufft(
            sinogram_ft.reshape(self.n, -1),
            pts.reshape(2, n_real_points * len(angles)),
            sz=(L, L),
            real=True,
        ).reshape(self.n, L, L)

        imgs = imgs / (self.n_radial_points * len(angles))
        return aspire.image.Image(xp.asnumpy(imgs)).stack_reshape(original_stack_shape)



import abc
import logging
from collections.abc import Iterable

import numpy as np

from aspire.basis import Basis, Coef, ComplexCoef
from aspire.operators import BlkDiagMatrix
from aspire.utils import LogFilterByCount, complex_type, real_type, trange

logger = logging.getLogger(__name__)


class SteerableBasis2D(Basis, abc.ABC):
    """
    `SteerableBasis2D` is an extension of Basis that is expected to have
    `rotation` (steerable) and `calculate_bispectrum` methods.
    """

    # Default matrix type for basis representation.
    matrix_type = BlkDiagMatrix

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Because they are used for core features of SteerableBasis2D,
        #   cache the indices for positive and negative ells.
        # Note zero is special case.
        self._zero_angular_inds = self.angular_indices == 0
        self._pos_angular_inds = (self.signs_indices == 1) & (self.angular_indices != 0)
        self._neg_angular_inds = self.signs_indices == -1
        self._non_neg_angular_inds = self.signs_indices >= 0
        self._blk_diag_cov_shape = None

        # Centralize indices attributes between FB/PSWF/FLE in SteerableBasis2D
        self.complex_count = self.count - sum(self._neg_angular_inds)
        self.complex_angular_indices = self.angular_indices[self._non_neg_angular_inds]
        self.complex_radial_indices = self.radial_indices[self._non_neg_angular_inds]

        # Attribute for caching the blk_diag shape once known.
        self._blk_diag_cov_shape = None

    def calculate_bispectrum(
        self, complex_coef, flatten=False, filter_nonzero_freqs=False, freq_cutoff=None
    ):
        """
        Calculate bispectrum for a set of coefs in this basis.

        The Bispectum matrix is of shape:
            (count, count, unique_radial_indices)

        where count is the number of complex coefficients.

        :param coef: Coefficients representing a (single) image expanded in this basis.
        :param flatten: Optionally extract symmetric values (tril) and then flatten.
        :param filter_nonzero_freqs: Remove indices corresponding to zero frequency (defaults False).
        :param freq_cutoff: Truncate (zero) high k frequecies above (int) value, defaults off (None).
        :return: Bispectum matrix (complex valued).
        """

        if not isinstance(complex_coef, Coef):
            raise TypeError(f"Expect `Coef` received {type(complex_coef)}.")
        complex_coef = complex_coef.asnumpy()

        # Check shape
        if complex_coef.shape[0] != 1:
            raise ValueError(
                "Due to potentially large sizes, bispectrum is limited to a single set of coefs."
                f"  Passed shape {complex_coef.shape}"
            )

        if complex_coef.shape[1] != self.complex_count:
            raise ValueError(
                "Basis.calculate_bispectrum coefs expected"
                f" to have (complex) count {self.complex_count}, received {complex_coef.shape}."
            )

        # From here just treat complex_coef as 1d vector instead of 1 by count 2d array.
        complex_coef = complex_coef[0]

        if freq_cutoff and freq_cutoff > np.max(self.complex_angular_indices):
            logger.warning(
                f"Bispectrum frequency cutoff {freq_cutoff} outside max {np.max(self.complex_angular_indices)}"
            )

        # Notes, regarding the naming:
        # radial freq indices q in paper/slides, _indices["ks"] in code
        radial_indices = self.complex_radial_indices  # q
        # angular freq indices k in paper/slides, _indices["ells"] in code
        angular_indices = self.complex_angular_indices  # k
        # Compute the set of all unique q in the compressed basis
        #   Note that np.unique is also sorted.
        unique_radial_indices = np.unique(radial_indices)

        # When compressed, we need maps between the basis and uncompressed set of q
        #   to a reduced set of q that remain after compression.
        # One map is provided by self.complex_radial_indices
        #   which maps an index in the basis remaining after compression to a q value.
        # The following code computes a similar but inverted map,
        #   given a q value, find an index into the set of unique q after compression.
        # Also, it is known that the set of q gets sparser with increasing k
        #   but that's ignored here, instead construct a dense
        #   array and filter it later.
        #  The plan is to revisit this code after appropriate coef classes are derived.

        # Default array to fill_value, we can use a value
        # k should never achieve..
        fill_value = self.complex_count**2
        compressed_radial_map = (
            np.ones(np.max(unique_radial_indices) + 1, dtype=int) * fill_value
        )
        for uniq_q_index, q_value in enumerate(unique_radial_indices):
            compressed_radial_map[q_value] = uniq_q_index

        B = np.zeros(
            (self.complex_count, self.complex_count, unique_radial_indices.shape[0]),
            dtype=complex_type(self.dtype),
        )

        logger.info(f"Calculating bispectrum matrix with shape {B.shape}.")

        for ind1 in range(self.complex_count):
            k1 = angular_indices[ind1]
            if freq_cutoff and k1 > freq_cutoff:
                continue
            coef1 = complex_coef[ind1]

            for ind2 in range(self.complex_count):
                k2 = angular_indices[ind2]
                if freq_cutoff and k2 > freq_cutoff:
                    continue
                coef2 = complex_coef[ind2]

                k3 = k1 + k2
                intermodulated_coef_inds = angular_indices == k3

                if np.any(intermodulated_coef_inds):
                    # Get the specific q indices related to feasible k3 angular_indices
                    Q3_ind = radial_indices[intermodulated_coef_inds]

                    if hasattr(self, "compressed") and self.compressed:
                        # Map those Q3_ind values to indices into compressed unique_radial_indices
                        #  by using the compressed_radial_map prepared above.
                        Q3_ind = compressed_radial_map[Q3_ind]

                    Coef3 = complex_coef[intermodulated_coef_inds]

                    B[ind1, ind2, Q3_ind] = coef1 * coef2 * np.conj(Coef3)

        if filter_nonzero_freqs:
            non_zero_freqs = angular_indices != 0
            B = B[non_zero_freqs][:, non_zero_freqs]

        if flatten:
            # B is sym, start by taking lower triangle.
            tril = np.tri(B.shape[0], dtype=bool)
            B = B[tril, :]
            # Then flatten
            B = B.flatten()

        return B

    def rotate(self, coef, radians, refl=None):
        """
        Returns coefs rotated counter-clockwise by `radians`.

        :param coef: Basis coefs.
        :param radians: Rotation in radians.
        :param refl: Optional reflect image (about y=0) (bool)
        :return: rotated coefs.
        """

        if not isinstance(coef, Coef):
            raise TypeError(f"`coef` must be `Coef` instance, received {type(coef)}.")

        coef = coef.asnumpy()

        # Covert radians to a broadcastable shape
        if isinstance(radians, Iterable):
            radians = np.array(radians, dtype=self.dtype)
            if radians.ndim < 2:
                radians = radians.reshape(-1, 1)
            else:
                radians = np.expand_dims(radians, axis=-1)

            if radians.size != np.prod(coef.shape[:-1]):
                raise RuntimeError(
                    f"`rotate` call `radians` {radians.shape} does not match"
                    f" `coef` {coef.shape[:-1]}."
                )
        # else: radians can be a constant

        assert (
            self.count == coef.shape[-1]
        ), "Number of coefficients must match self.count."

        # self.angular_indices are `ks`
        # For all coef in stack,
        #   compute the ks * radian used in the trig functions
        ks_rad = np.atleast_2d(self.angular_indices * radians)
        ks_pos = ks_rad[..., self._pos_angular_inds]
        ks_neg = ks_rad[..., self._neg_angular_inds]

        # Slice the coef on postive and negative ells
        coef_zer = coef[..., self._zero_angular_inds]
        coef_pos = coef[..., self._pos_angular_inds]
        coef_neg = coef[..., self._neg_angular_inds]

        # Handle zero case and avoid mutating the original array
        coef = np.empty_like(coef)
        coef[..., self._zero_angular_inds] = coef_zer

        # refl
        if refl is not None:
            if isinstance(refl, np.ndarray):
                assert len(refl) == len(coef)
            # else: refl can be a constant
            # negate the coefs corresponding to negative ells
            coef_neg[refl] = coef_neg[refl] * -1

        # Apply formula
        coef[..., self._pos_angular_inds] = coef_pos * np.cos(
            ks_pos
        ) + coef_neg * np.sin(ks_neg)
        coef[..., self._neg_angular_inds] = coef_neg * np.cos(
            ks_neg
        ) - coef_pos * np.sin(ks_pos)

        return Coef(self, coef)

    def complex_rotate(self, complex_coef, radians, refl=None):
        """
        Returns complex coefs rotated counter-clockwise by `radians`.

        This implementation uses the complex exponential.
        It is kept in the code for documentation and
        reference purposes.

        To invoke in code:

        self.to_real(
            self.complex_rotate(
                self.to_complex(coef), radians, refl)
            )
        )

        :param complex_coef: Basis coefs (in complex representation).
        :param radians: Rotation in radians.
        :param refl: Optional reflect image (about y=0) (bool)
        :return: rotated (complex) coefs.
        """

        # Covert radians to a broadcastable shape
        if isinstance(radians, np.ndarray):
            if len(radians) != len(complex_coef):
                raise RuntimeError(
                    "`rotate` call `radians` length cannot broadcast with"
                    f" `complex_coef` {len(complex_coef)} != {len(radians)}"
                )
            radians = radians.reshape(-1, 1)
        # else: radians can be a constant

        ks = self.complex_angular_indices
        assert len(ks) == complex_coef.shape[-1]

        # Don't mutate the input coef array (danger)
        _complex_coef = complex_coef.copy()

        # refl
        if refl is not None:
            if isinstance(refl, np.ndarray):
                assert len(refl) == len(complex_coef)
            # else: refl can be a constant
            # get the coefs corresponding to -ks , aka "ells"
            _complex_coef[refl] = np.conj(complex_coef[refl])

        _complex_coef = _complex_coef * np.exp(1j * ks * radians)

        return _complex_coef

    def shift(self, coef, shifts):
        """
        Returns coefs shifted by `shifts`.

        This will transform to real cartesian space, shift,
        and transform back to Polar Fourier-Bessel space.

        :param coef: Basis coefs.
        :param shifts: Shifts in pixels (x,y). Shape (1,2) or (len(coef), 2).
        :return: coefs of shifted images.
        """

        shifts = np.atleast_2d(np.array(shifts))
        if shifts.ndim != 2:
            raise ValueError("`shifts` should be a one or two dimensional array.")
        if shifts.shape[1] != 2 or shifts.shape[0] not in (1, len(coef)):
            raise ValueError(
                "`shifts` should be shape (1,2) or (len(coef),2),"
                f" received {shifts.shape}."
            )

        return self.evaluate_t(self.evaluate(coef).shift(shifts))

    @property
    def blk_diag_cov_shape(self):
        """
        Return the `BlkDiagMatrix` partition shapes.

        If the shape has already been cached,
        returns cached value.  Otherwise, will
        compute the shape and cache in this instance.
        """
        # Compute the _blk_diag_cov_shape as needed.
        if self._blk_diag_cov_shape is None:
            blks = []
            for ell in range(self.ell_max + 1):
                sgns = (1,) if ell == 0 else (1, -1)
                for _ in sgns:
                    blks.append(
                        [
                            self.k_max[ell],
                        ]
                        * 2
                    )
            self._blk_diag_cov_shape = np.array(blks)

        # Return the cached shape
        return self._blk_diag_cov_shape

    # This is included for completion, but is not being adopted yet.
    def indices_mask(self, **kwargs):
        """
        Given `radial=` or `angular=` expressions, return (`count`,)
        shaped mask where values satisfying the expression are `True`.

        Examples:
            No args yield all indices.
            `angular=0 creates a mask for selecting coefficients with zero angular indices.
            `angular=1, radial=2` selects coefficients satisfying angular index of 1 _and_ radial index of 2.
            More advanced operations can combine indices attributes.
             `angular=self.angular_indices>=0, radial=r` selects coefficients with non negative angular indices and some radial index `r`.

        :return: Boolen mask of shape (`count`,).
            Intended to be broadcast with `Coef` containers.
        """

        radial = kwargs.get("radial", None)
        angular = kwargs.get("angular", None)
        signs = kwargs.get("signs", None)

        # slowly construct the map
        signs_mask = np.zeros(self.count, dtype=bool)
        radial_mask = signs_mask.copy()
        angular_mask = signs_mask.copy()

        if radial is None:
            radial_mask[:] = True
        else:
            for k in np.atleast_1d(radial):
                radial_mask[self.radial_indices == k] = True

        if angular is None:
            angular_mask[:] = True
        else:
            for el in np.atleast_1d(angular):
                angular_mask[self.angular_indices == el] = True

        if signs is None:
            signs_mask[:] = True
        else:
            for s in np.atleast_1d(signs):
                signs_mask[self.signs_indices == s] = True

        mask = radial_mask & angular_mask & signs_mask

        return mask

    def to_real(self, complex_coef):
        """
        Return real valued representation of complex coefficients.
        This can be useful when comparing, prototyping, or
        implementing methods from literature.

        There is a corresponding method, `to_complex`.

        :param complex_coef: Complex `Coef` from this basis.
        :return: Real `Ceof` representation from this basis.
        """

        if not isinstance(complex_coef, ComplexCoef):
            raise TypeError(
                f"complex_coef should be instance of `Coef`, received {type(complex_coef)}."
            )

        if complex_coef.dtype not in (np.complex128, np.complex64):
            raise TypeError("coef provided to to_real should be complex.")

        # Pass through dtype precisions, but check and warn if mismatched.
        dtype = real_type(complex_coef.dtype)
        if dtype != self.dtype:
            logger.warning(
                f"Complex coef dtype {complex_coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
            )

        coef = np.zeros((*complex_coef.stack_shape, self.count), dtype=dtype)
        complex_coef = complex_coef.asnumpy()

        ind = 0
        idx = np.arange(self.k_max[0], dtype=int)
        ind += np.size(idx)
        ind_pos = ind

        coef[..., idx] = complex_coef[..., idx].real

        for ell in range(1, self.ell_max + 1):
            idx = ind + np.arange(self.k_max[ell], dtype=int)
            idx_pos = ind_pos + np.arange(self.k_max[ell], dtype=int)
            idx_neg = idx_pos + self.k_max[ell]

            c = complex_coef[..., idx]
            coef[..., idx_pos] = 2.0 * np.real(c)
            coef[..., idx_neg] = -2.0 * np.imag(c)

            ind += np.size(idx)
            ind_pos += 2 * self.k_max[ell]

        return Coef(self, coef)

    def to_complex(self, coef):
        """
        Return complex valued representation of complex coefficients.
        This can be useful when comparing, prototyping, or
        implementing methods from literature.

        There is a corresponding method, `to_real`.

        :param coef: Real `Coef` from this basis.
        :return: `ComplexCoef` representation from this basis.
        """

        if not isinstance(coef, Coef):
            raise TypeError(
                f"coef should be instance of `Coef`, received {type(coef)}."
            )

        if coef.dtype not in (np.float64, np.float32):
            raise TypeError("coef provided to to_complex should be real.")

        # Pass through dtype precions, but check and warn if mismatched.
        dtype = complex_type(coef.dtype)
        if coef.dtype != self.dtype:
            logger.warning(
                f"coef dtype {coef.dtype} does not match precision of basis.dtype {self.dtype}, returning {dtype}."
            )

        # Return the same precision as coef
        imaginary = dtype(1j)

        complex_coef = np.zeros((*coef.stack_shape, self.complex_count), dtype=dtype)
        coef = coef.asnumpy()

        ind = 0
        idx = np.arange(self.k_max[0], dtype=int)
        ind += np.size(idx)

        complex_coef[..., idx] = coef[..., idx]

        for ell in range(1, self.ell_max + 1):
            idx = ind + np.arange(self.k_max[ell], dtype=int)
            complex_coef[..., idx] = (
                coef[..., self._pos[idx]] - imaginary * coef[..., self._neg[idx]]
            ) / 2.0

            ind += np.size(idx)

        return ComplexCoef(self, complex_coef)

    # `abstractmethod` enforces when a new subclass of
    # `SteerableBasis2D` is created that this method is explicitly
    # implemented.  This is intended to encourage future basis authors
    # to consider this method for their application.
    @abc.abstractmethod
    def filter_to_basis_mat(self, f, method="evaluate_t", truncate=True):
        """
        Convert a filter into a basis operator representation.

        :param f: `Filter` object, usually a `CTFFilter`.
        :param method: `evaluate_t` or `expand`.
        :param truncate: Optionally, truncate dense matrix to BlkDiagMatrix.
            Defaults to True.

        :return: Representation of filter as `basis` operator.
            Return type will be based on the class's `matrix_type`.
        """
        # evaluate_t is not as accurate, but much much faster...
        if method == "evaluate_t":
            expand_method = self.evaluate_t
        elif method == "expand":
            expand_method = self.expand
        else:
            raise NotImplementedError(
                "`filter_to_basis_mat` method {method} not supported."
                "  Try `evaluate_t` or `expand`."
            )

        coef = Coef(self, np.eye(self.count, dtype=self.dtype))
        img = coef.evaluate()

        # Expansion can fail for some filters on specific basis vectors.
        # Loop over the expanding the filtered basis vectors one by one,
        # zero-ing failed vectors.
        filt = np.zeros((self.count, self.count), self.dtype)
        with LogFilterByCount(logger, 1):
            for i in trange(self.count):
                try:
                    filt[i] = expand_method(img[i].filter(f)).asnumpy()[0]
                except Exception:
                    logger.warning(
                        f"Failed to expand basis vector {i} after filter {f}."
                    )

        # Optionally truncate off block elements to zero.
        if truncate:
            filt = BlkDiagMatrix.from_dense(
                filt,
                self.blk_diag_cov_shape,
            )

        return filt



import logging
import os
import re
import sys
from glob import glob

logger = logging.getLogger(__name__)


def check_blank_line_above_param_section(file_path):
    """
    Check that every docstring with both a body section and a parameter
    section separates the two sections with exactly one blank line. Log
    errors and return count.

    :param file_path: File path to check for error.
    :return: Per file error count.
    """
    error_count = 0
    with open(file_path, "r") as file:
        content = file.read()

    regex = (
        r" {4,}\"\"\"\n(?:^[^:]+?[^\n])+(\n|\n\n\n+) {4,}(:p|:r)(?:.*\n)+? {4,}\"\"\""
    )

    bad_docstrings = re.finditer(regex, content, re.MULTILINE)
    for docstring in bad_docstrings:
        line_number = content.count("\n", 0, docstring.start()) + 1

        # Log error message.
        msg = "Must have exactly 1 blank line between docstring body and parameter sections."
        logger.error(f"{file_path}: {line_number}: {msg}")
        error_count += 1

    return error_count


def process_directory(directory):
    """
    Recursively walk through directories and check for docstring errors.
    If any errors found, log error count and exit.

    :param directory: Directory path to walk.
    """
    error_count = 0
    for file in glob(os.path.join(directory, "**/*.py"), recursive=True):
        error_count += check_blank_line_above_param_section(file)
    if error_count > 0:
        logger.error(f"Found {error_count} docstring errors.")
        sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        logger.warning("Usage: python check_docstrings.py <directory>")
        sys.exit(1)

    target_directory = sys.argv[1]
    if not os.path.isdir(target_directory):
        raise RuntimeError(f"Invalid target directory path: {target_directory}")
    process_directory(target_directory)



import logging
import os
from pathlib import Path

from aspire.basis import Coef, FFBBasis3D
from aspire.reconstruction.kernel import FourierKernel

logger = logging.getLogger(__name__)


class Estimator:
    def __init__(
        self,
        src,
        basis=None,
        batch_size=512,
        preconditioner="circulant",
        checkpoint_iterations=10,
        checkpoint_prefix="volume_checkpoint",
        maxiter=50,
        boost=True,
    ):
        """
        An object representing a 2*L-by-2*L-by-2*L array containing the non-centered Fourier transform of the mean
        least-squares estimator kernel.
        Convolving a volume with this kernel is equal to projecting and backproject-ing that volume in each of the
        projection directions (with the appropriate amplitude multipliers and CTFs) and averaging over the whole
        dataset.
        Note that this is a non-centered Fourier transform, so the zero frequency is found at index 1.

        :param src: `ImageSource` to be used for estimation.
        :param basis: 3D Basis to be used during estimation.
        :param batch_size: Optional batch size of images drawn from
            `src` during back projection and kernel estimation steps.
        :param preconditioner: Optional kernel preconditioner (`string`).
            Currently supported options are "circulant" or None.
        :param checkpoint_iterations: Optionally save `cg` estimated
            `basis` coefficients periodically each
            `checkpoint_iterations`.  Setting to `None` disables,
            otherwise checks for positive integer.  Note, when
            `maxiter` is not `None` and `cg` fails to converge a final
            checkpoint will still be written.
        :param checkpoint_prefix: Optional path prefix for `cg`
            checkpoint files.  If the parent directory does not exist,
            creation is attempted.  `_iter{N}` will be appended to the
            prefix.
        :param maxiter: Optional max number of `cg` iterations
            before returning.  This should be used in conjunction with
            `checkpoint_iterations` to prevent excessive disk usage.
            `None` disables.
        :param boost: Option to use `src` symmetry to boost number of images used for mean estimation (Boolean).
            Default of `True` employs symmetry boosting.
        """

        self.src = src
        if basis is None:
            logger.info(f"{self.__class__.__name__} instantiating default basis.")
            basis = FFBBasis3D(src.L, dtype=src.dtype)
        self.basis = basis
        self.dtype = self.src.dtype
        self.batch_size = batch_size
        if not preconditioner or preconditioner.lower() == "none":
            # Resolve None and string nones to None
            preconditioner = None
        elif preconditioner not in ["circulant"]:
            raise ValueError(
                f"Supplied preconditioner {preconditioner} is not supported."
            )
        self.preconditioner = preconditioner
        self.boost = boost

        # dtype configuration
        if not self.dtype == self.basis.dtype:
            logger.warning(
                f"Inconsistent types in {self.dtype} Estimator."
                f" basis: {self.basis.dtype}"
            )

        if src.L != basis.nres:
            raise ValueError(
                "Currently require 2D source and 3D volume resolution to be the same."
                f" Given src.L={src.L} != {basis.nres}"
            )

        # Checkpoint configuration
        if checkpoint_iterations is not None:
            try:
                checkpoint_iterations = int(checkpoint_iterations)
            except ValueError:
                # Sentinel value to emit a more descriptive message below.
                checkpoint_iterations = -1

            if not checkpoint_iterations > 0:
                raise ValueError(
                    "`checkpoint_iterations` should be a positive integer or `None`."
                )
        self.checkpoint_iterations = checkpoint_iterations

        # Create checkpointing dirs as needed
        if checkpoint_prefix:
            parent = Path(checkpoint_prefix).parent
            if not os.path.exists(parent):
                os.makedirs(parent)
        self.checkpoint_prefix = checkpoint_prefix

        # Maximum iteration configuration
        if maxiter is not None:
            try:
                maxiter = int(maxiter)
            except ValueError:
                # Sentinel value to emit a more descriptive message below.
                maxiter = -1
            if not maxiter > 0:
                raise ValueError("`maxiter` should be a positive integer or `None`.")
        self.maxiter = maxiter

    def __getattr__(self, name):
        """Lazy attributes instantiated on first-access"""

        if name == "kernel":
            logger.info("Computing kernel")
            kernel = self.kernel = self.compute_kernel()
            return kernel

        elif name == "precond_kernel":
            if self.preconditioner == "circulant":
                logger.info("Computing Preconditioner kernel")
                precond_kernel = self.precond_kernel = FourierKernel(
                    1.0 / self.kernel.circularize()
                )
            else:
                precond_kernel = self.precond_kernel = None
            return precond_kernel

        else:
            raise AttributeError(name)

    def compute_kernel(self):
        raise NotImplementedError("Subclasses must implement the compute_kernel method")

    def estimate(self, b_coef=None, x0=None, tol=1e-5, regularizer=0):
        """Return an estimate as a Volume instance."""
        if b_coef is None:
            b_coef = self.src_backward()
        est_coef = self.conj_grad(b_coef, x0=x0, tol=tol, regularizer=regularizer)
        est = Coef(self.basis, est_coef).evaluate()

        return est

    def apply_kernel(self, vol_coef, kernel=None):
        """
        Applies the kernel represented by convolution

        :param vol_coef: The volume to be convolved, stored in the basis coefficients.
        :param kernel: a Kernel object. If None, the kernel for this Estimator is used.
        :return: The result of evaluating `vol_coef` in the given basis, convolving with the kernel given by
            kernel, and backprojecting into the basis.
        """

        if kernel is None:
            kernel = self.kernel

        vol = Coef(self.basis, vol_coef).evaluate()  # returns a Volume
        vol = kernel.convolve_volume(vol)  # returns a Volume
        vol_coef = self.basis.evaluate_t(vol)
        return vol_coef



import logging
import os.path
import warnings

import numpy as np
from numpy.linalg import norm
from scipy.optimize import curve_fit

from aspire.abinitio import CLOrient3D, SyncVotingMixin
from aspire.utils import (
    J_conjugate,
    Rotation,
    all_pairs,
    nearest_rotations,
    tqdm,
    trange,
)
from aspire.utils.matlab_compat import stable_eigsh
from aspire.utils.random import rand

logger = logging.getLogger(__name__)


class CLSync3N(CLOrient3D, SyncVotingMixin):
    """
    Define a class to estimate 3D orientations using common lines Sync3N methods (2017).

    Ido Greenberg, Yoel Shkolnisky,
    Common lines modeling for reference free Ab-initio reconstruction in cryo-EM,
    Journal of Structural Biology,
    Volume 200, Issue 2,
    2017,
    Pages 106-117,
    ISSN 1047-8477,
    https://doi.org/10.1016/j.jsb.2017.09.007.
    """

    # Initialize alternatives
    #
    # When we find the best J-configuration, we also compare it to the alternative 2nd best one.
    # this comparison is done for every pair in the triplet independently. to make sure that the
    # alternative is indeed different in relation to the pair, we document the differences between
    # the configurations in advance:
    # ALTS(:,best_conf,pair) = the two configurations in which J-sync differs from best_conf in relation to pair

    _ALTS = np.array(
        [
            [[1, 2, 1], [0, 2, 0], [0, 0, 1], [1, 0, 0]],
            [[2, 3, 3], [3, 3, 2], [3, 1, 3], [2, 1, 2]],
        ],
        dtype=int,
    )

    def __init__(
        self,
        src,
        n_rad=None,
        n_theta=360,
        max_shift=0.15,
        shift_step=1,
        hist_bin_width=1,
        full_width="adaptive",
        epsilon=1e-2,
        max_iters=1000,
        seed=None,
        mask=True,
        S_weighting=False,
        J_weighting=False,
        hist_intervals=100,
        disable_gpu=False,
    ):
        """
        Initialize object for estimating 3D orientations.

        :param src: The source object of 2D denoised or class-averaged images with metadata
        :param n_rad: The number of points in the radial direction
        :param n_theta: The number of points in the theta direction
        :param max_shift: Maximum range for shifts as a proportion of box size. Default = 0.15.
        :param shift_step: Step size of shift estimation in pixels. Default = 1 pixel.
        :param hist_bin_width: Bin width in smoothing histogram (degrees).
        :param full_width: Selection width around smoothed histogram peak (degrees).
            `adaptive` will attempt to automatically find the smallest number of
            `hist_bin_width`s required to find at least one valid image index.
        :param epsilon: Tolerance for the power method.
        :param max_iter: Maximum iterations for the power method.
        :param seed: Optional seed for RNG.
        :param mask: Option to mask `src.images` with a fuzzy mask (boolean).
            Default, `True`, applies a mask.
        :param S_weighting: Optionally apply probabilistic weighting
            to the `S` matrix.
        :param J_weighting: Optionally use `J` weights instead of
            signs when computing `signs_times_v`.
        :param hist_intervals: Number of histogram bins used to
            compute triangle scores when `S_weighting` enabled.
        :param disable_gpu: Disables GPU acceleration;
            forces CPU only code for this module.
            Defaults to automatically using GPU when available.
        """

        super().__init__(
            src,
            n_rad=n_rad,
            n_theta=n_theta,
            max_shift=max_shift,
            shift_step=shift_step,
            hist_bin_width=hist_bin_width,
            full_width=full_width,
            mask=mask,
        )

        # Generate pair mappings
        self._pairs, self._pairs_to_linear = all_pairs(self.n_img, return_map=True)

        self.epsilon = epsilon
        self.max_iters = max_iters
        self.seed = seed

        # Sync3N specific vars
        self.S_weighting = S_weighting
        self.J_weighting = J_weighting
        self._D_null = 1e-13
        self.hist_intervals = int(hist_intervals)
        # Warn if histogram may be too sparse for curve fitting
        if self.S_weighting and (src.n < hist_intervals):
            logger.warning(
                f"`hist_intervals` {hist_intervals} > src.n {src.n}."
                "  Consider reducing if curve fitting is infeasable."
            )

        # Auto configure GPU
        self.__gpu_module = None
        if not disable_gpu:
            try:
                import cupy as cp

                if cp.cuda.runtime.getDeviceCount() >= 1:
                    gpu_id = cp.cuda.runtime.getDevice()
                    logger.info(
                        f"cupy and GPU {gpu_id} found by cuda runtime; enabling cupy."
                    )
                    self.__gpu_module = self.__init_cupy_module()
                else:
                    logger.info("GPU not found, defaulting to numpy.")

            except ModuleNotFoundError:
                logger.info("cupy not found, defaulting to numpy.")

    ###########################################
    # High level algorithm steps              #
    ###########################################
    def estimate_rotations(self):
        """
        Estimate rotation matrices.

        :return: Array of rotation matrices, size n_imgx3x3.
        """

        logger.info(f"Estimating relative viewing directions for {self.n_img} images.")

        # Detect a single pair of common-lines between each pair of images
        self.build_clmatrix()

        # Initial estimate of viewing directions
        # Calculate relative rotations
        Rijs0 = self._estimate_all_Rijs(self.clmatrix)

        # Compute and apply global handedness
        Rijs = self._global_J_sync(Rijs0)

        # Build sync3n matrix
        S = self._construct_sync3n_matrix(Rijs)

        # Optionally compute S weights
        W = None
        if self.S_weighting is True:
            W = self._syncmatrix_weights(Rijs)

        # Yield rotations from S
        self.rotations = self._sync3n_S_to_rot(S, W)

    #######################
    # Main Sync3N Methods #
    #######################
    def _sync3n_S_to_rot(self, S, W=None, n_eigs=4):
        """
        Use eigen decomposition of S to estimate transforms,
        then project transforms to nearest rotations.

        :param S: Numpy array representing Synchronization matrix.
        :param W: Optional weights array, default `None` is equal weighting of `S`.
        :param n_eigs: Optional, number of eigenvalues to compute (min 3).
        """

        # Critical this occurs in double precision
        S = S.astype(np.float64, copy=False)

        if n_eigs < 3:
            raise ValueError(
                f"n_eigs must be greater than 3, default is 4. Invoked with {n_eigs}"
            )

        if W is not None:
            logger.info("Applying weights to synchronization matrix.")
            if not W.shape == (self.n_img, self.n_img):
                raise RuntimeError(
                    f"Shape of W should be {(self.n_img, self.n_img)}."
                    f" Received {W.shape}."
                )
            # Initialize D
            # Critical this occurs in double precision
            W = W.astype(np.float64, copy=False)
            D = np.mean(W, axis=1)

            Dhalf = D
            # Compute mask of trouble D values
            nulls = np.abs(D) < self._D_null
            # Avoid trouble values when exponentiating
            Dhalf[~nulls] = Dhalf[~nulls] ** (-0.5)
            # Flush trouble values to zero
            Dhalf[nulls] = 0
            # expand diagonal
            Dhalf = np.diag(Dhalf)

            # Report W Diagnostic
            W_normalized = Dhalf**2 @ W
            nzidx = np.sum(W_normalized, axis=1) != 0
            err = np.linalg.norm(np.sum(W_normalized[nzidx], axis=1) - self.n_img)
            if err > 1e-10:
                logger.warning(f"Large Weights Matrix Normalization Error: {err}")

            # Make W of size 3Nx3N
            W = np.kron(W, np.ones((3, 3), dtype=self.dtype))

            # Make Dhalf of size 3Nx3N
            Dhalf = np.diag(np.kron(np.diag(Dhalf), np.ones(3, dtype=np.float64)))

            # Apply weights to S
            S = Dhalf @ (W * S) @ Dhalf

        # Extract three eigenvectors corresponding to non-zero eigenvalues.
        d, v = stable_eigsh(S, n_eigs, which="LM")

        sort_idx = np.argsort(-d)
        logger.info(
            f"Top {n_eigs} eigenvalues from synchronization voting matrix: {d[sort_idx]}"
        )

        # Only need the top 3 eigen-vectors.
        v = v[:, sort_idx[:3]]

        # Cancel symmetrization when using weights W
        if W is not None:
            # Untill now we used a symmetrized variant of the weighted Sync matrix,
            # thus we didn't get the right eigenvectors. to fix that we just need
            # to multiply:
            v = Dhalf @ v

        # Yield estimated rotations from the eigen-vectors
        rotations = v.reshape(self.n_img, 3, 3).transpose(0, 2, 1)

        # Enforce we are returning actual rotations
        rotations = nearest_rotations(rotations, allow_reflection=True)

        return rotations.astype(self.dtype)

    def _construct_sync3n_matrix(self, Rij):
        """
        Construct sync3n matrix from estimated rotations Rij.

        :param Rij: Numpy array of estimated rotations (all pairs).
        :return: Synchronization matrix S, (3*N, 3*N).
        """

        # Initialize S with diag identity blocks
        n = self.n_img
        S = np.eye(3 * n, dtype=self.dtype).reshape(n, 3, n, 3)

        idx = 0
        for i in range(n):
            for j in range(i + 1, n):
                # S( (3*i-2):(3*i) , (3*j-2):(3*j) ) = Rij(:,:,idx); % Rij
                S[i, :, j, :] = Rij[idx]
                # S( (3*j-2):(3*j) , (3*i-2):(3*i) ) = Rij(:,:,idx)'; % Rji = Rij'
                S[j, :, i, :] = Rij[idx].T
                idx += 1

        # Convert S shape to 3Nx3N
        S = S.reshape(3 * n, 3 * n)

        return S

    def _syncmatrix_weights(
        self,
        Rijs,
        permitted_inconsistency=1.5,
        p_domain_limit=0.7,
        max_iterations=12,
        min_p_permitted=0.04,
    ):
        """
        Given relative rotations matrix `Rij`,
        compute and return probability weights `W` for S.

        Default parameters here were taken from those in the MATLAB
        code, with the original author noting they were found
        empirically.

        :param permitted_inconsistency: Consistency condition is
            `mean(Pij)/permitted_inconsistency < P <
            mean(Pij)*permitted_inconsistency`.
        :param p_domain_limit: Domain of P is [Pmin,Pmax], with
            Pmin=p_domain_limit*Pmax
        :param max_iterations: Maximum iterations for P estimation.
        :param min_p_permitted: Small value at which to stop
            attempting to synchronize P.
        :return: Synchronization matrix weights `W`.
        """
        logger.info("Computing synchronization matrix weights.")

        def _body(prev_too_low, Pmin, Pmax, hist, p_domain_limit=p_domain_limit):
            """
            Helper function to run and test triangle_scores.
            """
            # Get inistial estimate for Pij
            P, sigma, Pij, hist = self._triangle_scores(Rijs, hist, Pmin, Pmax)

            # Check if P and Pij are consistent
            mean_Pij = np.mean(Pij)
            too_low = P < mean_Pij / permitted_inconsistency
            too_high = P > mean_Pij * permitted_inconsistency
            inconsistent = too_low | too_high

            # Check trend
            if prev_too_low is not None and too_low != prev_too_low:
                p_domain_limit = np.sqrt(p_domain_limit)

            # define limits for next P estimation
            if too_high:
                if P < min_p_permitted:
                    logger.error(
                        "Triangles Scores are poorly distributed, whatever small P we force."
                    )

                if Pmax is not None:
                    Pmax = Pmax * p_domain_limit
                else:
                    Pmax = P

                Pmin = Pmax * p_domain_limit
            else:  # too low
                if Pmin is not None:
                    Pmin = Pmin / p_domain_limit
                else:
                    Pmin = P

                Pmax = Pmin / p_domain_limit

            return inconsistent, Pij, (too_low, Pmin, Pmax, hist)

        # Repeat iteratively until estimations of P & Pij are consistent
        i = 0
        res = (None,) * 4
        inconsistent = True
        while inconsistent and i < max_iterations:
            inconsistent, Pij, res = _body(*res)
            i += 1

        # Pack W
        W = np.zeros((self.n_img, self.n_img), dtype=self.dtype)
        idx = 0
        for i in range(self.n_img):
            for j in range(i + 1, self.n_img):
                W[i, j] = Pij[idx]
                W[j, i] = Pij[idx]
                idx += 1

        return W

    def _triangle_scores_inner(self, Rijs):
        """
        Computes histogram of `triangle scores`.

        Wrapper for cpu/gpu dispatch.

        :param Rijs: nchoose2 by 3 by 3 array of rotations.
        :return: Histogram of triangle scores.
        """

        # host/gpu dispatch
        if self.__gpu_module:
            scores_hist = self._triangle_scores_inner_cupy(Rijs)
        else:
            scores_hist = self._triangle_scores_inner_host(Rijs)

        return scores_hist

    def _triangle_scores_inner_host(self, Rijs):
        """
        See _triangle_scores_inner.

        CPU implementation.
        """

        # The following is adopted from Matlab triangle_scores_mex.c

        # Initialize probability result arrays
        scores_hist = np.zeros(self.hist_intervals, dtype=np.uint32)

        c = np.empty((4), dtype=Rijs.dtype)
        s = np.empty((3), dtype=Rijs.dtype)
        for i in trange(self.n_img - 2, desc="Computing triangle scores"):
            for j in range(
                i + 1, self.n_img - 1
            ):  # check bound (taken from MATLAB mex)
                ij = self._pairs_to_linear[i, j]
                Rij = Rijs[ij]
                for k in range(j + 1, self.n_img):
                    ik = self._pairs_to_linear[i, k]
                    jk = self._pairs_to_linear[j, k]
                    Rik = Rijs[ik]
                    Rjk = Rijs[jk]

                    # Compute conjugated rotats
                    Rij_J = J_conjugate(Rij)
                    Rik_J = J_conjugate(Rik)
                    Rjk_J = J_conjugate(Rjk)

                    # Compute R muls and norms
                    c[0] = np.sum(((Rij @ Rjk) - Rik) ** 2)
                    c[1] = np.sum(((Rij_J @ Rjk) - Rik) ** 2)
                    c[2] = np.sum(((Rij @ Rjk_J) - Rik) ** 2)
                    c[3] = np.sum(((Rij @ Rjk) - Rik_J) ** 2)

                    # Find best match
                    best_i = np.argmin(c)
                    best_val = c[best_i]

                    # For each triangle side, find the best alternative
                    alt_ij_jk = c[self._ALTS[0][best_i][0]]
                    if c[self._ALTS[1][best_i][0]] < alt_ij_jk:
                        alt_ij_jk = c[self._ALTS[1][best_i][0]]

                    alt_ik_jk = c[self._ALTS[0][best_i][1]]
                    if c[self._ALTS[1][best_i][1]] < alt_ik_jk:
                        alt_ik_jk = c[self._ALTS[1][best_i][1]]

                    alt_ij_ik = c[self._ALTS[0][best_i][2]]
                    if c[self._ALTS[1][best_i][2]] < alt_ij_ik:
                        alt_ij_ik = c[self._ALTS[1][best_i][2]]

                    # Compute scores
                    s[0] = 1 - np.sqrt(best_val / alt_ij_jk)  # s_ij_jk
                    s[1] = 1 - np.sqrt(best_val / alt_ik_jk)  # s_ik_jk
                    s[2] = 1 - np.sqrt(best_val / alt_ij_ik)  # s_ij_ik

                    # Update histogram
                    # Find integer bin [0,self.hist_intervals)
                    _l1, _l2, _l3 = np.maximum(
                        np.minimum(
                            (self.hist_intervals * s).astype(int),  # implicit floor
                            self.hist_intervals - 1,  # clamp upper bound
                        ),
                        0,  # clamp lower bound
                    )

                    scores_hist[_l1] += 1
                    scores_hist[_l2] += 1
                    scores_hist[_l3] += 1

        return scores_hist

    def _triangle_scores_inner_cupy(self, Rijs):
        """
        See _triangle_scores_inner.

        GPU implementation.
        """

        import cupy as cp

        triangle_scores = self.__gpu_module.get_function("triangle_scores_inner")

        Rijs_dev = cp.array(Rijs, dtype=np.float64)

        # This holds integer counts
        scores_hist_dev = cp.zeros((self.hist_intervals), dtype=np.uint32)

        # call the kernel
        blkszx = 512
        nblkx = (self.n_img + blkszx - 1) // blkszx
        triangle_scores(
            (nblkx,),
            (blkszx,),
            (
                self.n_img,
                Rijs_dev,
                self.hist_intervals,
                scores_hist_dev,
            ),
        )

        # d2h
        scores_hist = scores_hist_dev.get()

        return scores_hist

    def _pairs_probabilities(self, Rijs, P2, A, a, B, b, x0):
        """
        This function computes the probability of a pair `ij` having
        an observed value of triangles score under two priors.  Once
        given it has an indicative common line, and again once given
        it has an arbitrary common line.

        The probability of the common line to be indicative can then
        be derived by Bayes Theorem.

        Wrapper for cpu/gpu dispatch.

        :param Rijs: nchoose2 by 3 by 3 array of rotations.
        :param P2: distribution parameter
        :param A: distribution parameter
        :param a: distribution parameter
        :param B: distribution parameter
        :param b: distribution parameter
        :param x0: Initial guess
        :return: (log indicative probabilities, log arbitrary probabilities)
        """
        # These param values are passed to C, force doubles.
        params = np.array([P2, A, a, B, b, x0], dtype=np.float64)

        # host/gpu dispatch
        if self.__gpu_module:
            ln_f_ind, ln_f_arb = self._pairs_probabilities_cupy(Rijs, *params)
        else:
            ln_f_ind, ln_f_arb = self._pairs_probabilities_host(Rijs, *params)

        return ln_f_ind, ln_f_arb

    def _pairs_probabilities_host(self, Rijs, P2, A, a, B, b, x0):
        """
        See _pairs_probabilities.

        CPU implementation.
        """
        # The following is adopted from Matlab pairs_probabilities_mex.c `looper`

        # Initialize probability result arrays
        ln_f_ind = np.zeros(len(Rijs), dtype=Rijs.dtype)
        ln_f_arb = np.zeros(len(Rijs), dtype=Rijs.dtype)

        c = np.empty((4), dtype=Rijs.dtype)
        for i in trange(self.n_img - 2, desc="Computing pair probabilities"):
            for j in range(i + 1, self.n_img - 1):
                ij = self._pairs_to_linear[i, j]
                Rij = Rijs[ij]
                for k in range(j + 1, self.n_img):
                    ik = self._pairs_to_linear[i, k]
                    jk = self._pairs_to_linear[j, k]
                    Rik = Rijs[ik]
                    Rjk = Rijs[jk]

                    # Compute conjugated rotats
                    Rij_J = J_conjugate(Rij)
                    Rik_J = J_conjugate(Rik)
                    Rjk_J = J_conjugate(Rjk)

                    # Compute R muls and norms
                    c[0] = np.sum(((Rij @ Rjk) - Rik) ** 2)
                    c[1] = np.sum(((Rij_J @ Rjk) - Rik) ** 2)
                    c[2] = np.sum(((Rij @ Rjk_J) - Rik) ** 2)
                    c[3] = np.sum(((Rij @ Rjk) - Rik_J) ** 2)

                    # Find best match
                    best_i = np.argmin(c)
                    best_val = c[best_i]

                    # For each triangle side, find the best alternative
                    alt_ij_jk = c[self._ALTS[0][best_i][0]]
                    if c[self._ALTS[1][best_i][0]] < alt_ij_jk:
                        alt_ij_jk = c[self._ALTS[1][best_i][0]]
                    alt_ik_jk = c[self._ALTS[0][best_i][1]]
                    if c[self._ALTS[1][best_i][1]] < alt_ik_jk:
                        alt_ik_jk = c[self._ALTS[1][best_i][1]]
                    alt_ij_ik = c[self._ALTS[0][best_i][2]]
                    if c[self._ALTS[1][best_i][2]] < alt_ij_ik:
                        alt_ij_ik = c[self._ALTS[1][best_i][2]]

                    # Compute scores
                    s_ij_jk = 1 - np.sqrt(best_val / alt_ij_jk)
                    s_ik_jk = 1 - np.sqrt(best_val / alt_ik_jk)
                    s_ij_ik = 1 - np.sqrt(best_val / alt_ij_ik)

                    # Update probabilities
                    # # Probability of pair ij having score given indicicative common line
                    # P2, B, b, x0, A, a
                    f_ij_jk = np.log(
                        P2
                        * (
                            B
                            * np.power(1 - s_ij_jk, b)
                            * np.exp(-b / (1 - x0) * (1 - s_ij_jk))
                        )
                        + (1 - P2) * A * np.power((1 - s_ij_jk), a)
                    )
                    f_ik_jk = np.log(
                        P2
                        * (
                            B
                            * np.power(1 - s_ik_jk, b)
                            * np.exp(-b / (1 - x0) * (1 - s_ik_jk))
                        )
                        + (1 - P2) * A * np.power((1 - s_ik_jk), a)
                    )
                    f_ij_ik = np.log(
                        P2
                        * (
                            B
                            * np.power(1 - s_ij_ik, b)
                            * np.exp(-b / (1 - x0) * (1 - s_ij_ik))
                        )
                        + (1 - P2) * A * np.power((1 - s_ij_ik), a)
                    )
                    ln_f_ind[ij] += f_ij_jk + f_ij_ik
                    ln_f_ind[jk] += f_ij_jk + f_ik_jk
                    ln_f_ind[ik] += f_ik_jk + f_ij_ik

                    # # Probability of pair ij having score given arbitrary common line
                    f_ij_jk = np.log(A * np.power((1 - s_ij_jk), a))
                    f_ik_jk = np.log(A * np.power((1 - s_ik_jk), a))
                    f_ij_ik = np.log(A * np.power((1 - s_ij_ik), a))
                    ln_f_arb[ij] += f_ij_jk + f_ij_ik
                    ln_f_arb[jk] += f_ij_jk + f_ik_jk
                    ln_f_arb[ik] += f_ik_jk + f_ij_ik

        return ln_f_ind, ln_f_arb

    def _pairs_probabilities_cupy(self, Rijs, P2, A, a, B, b, x0):
        """
        See _pairs_probabilities.

        GPU implementation.
        """

        import cupy as cp

        pairs_probabilities = self.__gpu_module.get_function("pairs_probabilities")

        Rijs_dev = cp.array(Rijs, dtype=np.float64)
        ln_f_ind_dev = cp.zeros((self.n_img * (self.n_img - 1) // 2), dtype=np.float64)
        ln_f_arb_dev = cp.zeros((self.n_img * (self.n_img - 1) // 2), dtype=np.float64)

        # call the kernel
        blkszx = 512
        nblkx = (self.n_img + blkszx - 1) // blkszx
        pairs_probabilities(
            (nblkx,),
            (blkszx,),
            (self.n_img, Rijs_dev, P2, A, a, B, b, x0, ln_f_ind_dev, ln_f_arb_dev),
        )

        # accumulate over thread results
        ln_f_arb = ln_f_arb_dev.get().astype(self.dtype, copy=False)
        ln_f_ind = ln_f_ind_dev.get().astype(self.dtype, copy=False)

        return ln_f_ind, ln_f_arb

    def _triangle_scores(
        self,
        Rijs,
        scores_hist,
        Pmin,
        Pmax,
        a=2.2,
        peak2sigma=2.43e-2,
        P=0.5,
        b=2.5,
        x0=0.78,
    ):
        """
        Computes `triangle_scores`, attempts to fit curve to
        distribution, and uses estimated distribution to compute
        `pairs_probabilities`.

        Default parameters here were taken from those in the MATLAB
        code, with the original author noting they were found
        empirically.

        :param a: distribution parameter
        :param peak2sigma: empirical relation between the location of
            the peak of the histigram, and the mean error in the
            common lines estimations.
        :param P: distribution parameter
        :param b: distribution parameter
        :param x0: Initial guess
        :return: Tuple of pairs probabilty Pij and related terms
             (P, sigma, Pij, scores_hist)
        """

        Pmin = Pmin or 0
        Pmin = max(Pmin, 0)  # Clamp probability to [0,1]
        Pmax = Pmax or 1
        Pmax = min(Pmax, 1)  # Clamp probability to [0,1]

        if scores_hist is None:
            scores_hist = self._triangle_scores_inner(Rijs)

        # Histogram decomposition: P & sigma evaluation
        h = 1 / self.hist_intervals
        hist_x = np.arange(h / 2, 1, h)
        # normalization factor of one component of the histogram
        A = (
            (self.n_img * (self.n_img - 1) * (self.n_img - 2) / 2)
            / self.hist_intervals
            * (a + 1)
        )
        # normalization of 2nd component: B = P*N_delta/sum(f), where f is the component formula
        # B0 = (
        #     P
        #     * (self.n_img * (self.n_img - 1) * (self.n_img - 2) / 2)
        #     / np.sum(((1 - hist_x) ** b) * np.exp(-b / (1 - x0) * (1 - hist_x)))
        # )
        # P must be in lower and upper bounds or `curve_fit` will error
        # This was not the case for MATLAB...
        # P0 = np.clip(P, Pmin**3, Pmax**3)
        # Note, MATLAB suggests the following, but I feel it is a bug.
        # Will discuss with Yoel about the original code's intent.
        #     np.array([B0, P0, b, x0], dtype=np.float64)
        start_values = None
        lower_bounds = np.array([0, Pmin**3, 2, 0], dtype=np.float64)
        upper_bounds = np.array([np.inf, Pmax**3, np.inf, 1], dtype=np.float64)

        with np.printoptions(precision=2):
            logger.info(f"curve_fit lower_bounds:{lower_bounds}")
            logger.info(f"curve_fit start_values:{start_values}")
            logger.info(f"curve_fit upper_bounds:{upper_bounds}")

        # Fit distribution
        def fun(x, B, P, b, x0, A=A, a=a):
            """Function to fit. x is data vector."""
            return (1 - P) * A * (1 - x) ** a + P * B * (1 - x) ** b * np.exp(
                -b / (1 - x0) * (1 - x)
            )

        popt, pcov = curve_fit(
            fun,
            hist_x.astype(np.float64, copy=False),
            scores_hist.astype(np.float64, copy=False),
            p0=start_values,
            bounds=(lower_bounds, upper_bounds),
            method="trf",  # MATLAB used method "LAR" with algo "Trust-Region"
        )
        B, P, b, x0 = popt

        # Derive P and sigma
        P = P ** (1 / 3)
        sigma = (1 - x0) / peak2sigma

        logger.info(f"Estimated CL Errors P,STD:\t{100*P:.2f}%\t{sigma:.2f}")

        # Initialize probability computations
        # Local histograms analysis
        A = a + 1  # distribution 1st component normalization factor
        # distribution 2nd component normalization factor
        B = B / (
            (self.n_img * (self.n_img - 1) * (self.n_img - 2) / 2) / self.hist_intervals
        )

        # Calculate probabilities
        ln_f_ind, ln_f_arb = self._pairs_probabilities(Rijs, P**2, A, a, B, b, x0)

        with warnings.catch_warnings():
            # For large values of (ln_f_arb - ln_f_ind), numpy exponential will overflow. We still
            # get the intended result of Pij = 0, so we capture and ignore the overflow warning.
            warnings.filterwarnings("ignore", r".*overflow encountered in exp.*")

            Pij = 1 / (1 + (1 - P) / P * np.exp(ln_f_arb - ln_f_ind))

        # Fix singular output
        num_nan = np.sum(np.isnan(Pij))
        if num_nan > 0:
            logger.error(
                f"NaN probabilities occurred {num_nan} times out of {np.size(Pij)}. Setting NaNs to zero."
            )
            Pij = np.nan_to_num(Pij)

        logger.info(
            f"Common lines probabilities to be indicative Pij={100*np.mean(Pij):.2f}%"
        )

        return P, sigma, Pij, scores_hist

    ###########################################
    # Primary Methods                         #
    ###########################################

    def _global_J_sync(self, Rijs):
        """
        Apply global J-synchronization.

        Given all pairs of estimated rotation matrices `Rijs` with
        arbitrary handedness (J conjugation), attempt to detect and
        conjugate entries of `Rijs` such that all rotations have same
        handedness.

        :param Rijs: Array of all pairs of rotation matrices
        :return: Array of all pairs of J synchronized rotation matrices
        """

        # Determine relative handedness of Rijs.
        sign_ij_J = self._J_sync_power_method(Rijs)

        # Synchronize Rijs
        logger.info("Applying global handedness synchronization.")
        mask = sign_ij_J == -1
        Rijs[mask] = J_conjugate(Rijs[mask])

        return Rijs

    def _estimate_all_Rijs(self, clmatrix):
        """
        Estimate Rijs using the voting method.

        :param clmatrix: Common lines matrix
        :return: Estimated rotations
        """
        n_img = self.n_img
        n_theta = self.n_theta
        Rijs = np.zeros((len(self._pairs), 3, 3))

        for idx, (i, j) in enumerate(tqdm(self._pairs, desc="Estimate Rijs")):
            Rijs[idx] = self._syncmatrix_ij_vote_3n(
                clmatrix, i, j, np.arange(n_img), n_theta
            )

        return Rijs

    def _syncmatrix_ij_vote_3n(self, clmatrix, i, j, k_list, n_theta):
        """
        Compute the (i,j) rotation block of the synchronization matrix using voting method

        Given the common lines matrix `clmatrix`, a list of images specified in k_list
        and the number of common lines n_theta, find the (i, j) rotation block Rij.

        :param clmatrix: The common lines matrix
        :param i: The i image
        :param j: The j image
        :param k_list: The list of images for the third image for voting algorithm
        :param n_theta: The number of points in the theta direction (common lines)
        :return: The (i,j) rotation block of the synchronization matrix
        """
        alphas, good_k = self._vote_ij(clmatrix, n_theta, i, j, k_list, sync=True)

        angles = np.zeros(3)

        if alphas is not None:
            angles[0] = clmatrix[i, j] * 2 * np.pi / n_theta + np.pi / 2
            angles[1] = np.mean(alphas)
            angles[2] = -np.pi / 2 - clmatrix[j, i] * 2 * np.pi / n_theta
            rot = Rotation.from_euler(angles).matrices

        else:
            # This is for the case that images i and j correspond to the same
            # viewing direction and differ only by in-plane rotation.
            # We set to zero as in the Matlab code.
            rot = np.zeros((3, 3))

        return rot

    #######################################
    # Secondary Methods for Global J Sync #
    #######################################

    def _J_sync_power_method(self, Rijs):
        """
        Calculate the leading eigenvector of the J-synchronization matrix
        using the power method.

        As the J-synchronization matrix is of size (n-choose-2)x(n-choose-2), we
        use the power method to compute the eigenvalues and eigenvectors,
        while constructing the matrix on-the-fly.

        :param Rijs: (n-choose-2)x3x3 array of estimates of relative orientation matrices.

        :return: An array of length n-choose-2 consisting of 1 or -1, where the sign of the
        i'th entry indicates whether the i'th relative orientation matrix will be J-conjugated.
        """

        logger.info(
            "Initiating power method to estimate J-synchronization matrix eigenvector."
        )
        # Set power method tolerance and maximum iterations.
        epsilon = self.epsilon
        max_iters = self.max_iters

        # Initialize candidate eigenvectors
        n_Rijs = Rijs.shape[0]
        vec = rand(n_Rijs, seed=self.seed)
        vec = vec / norm(vec)
        residual = 1
        itr = 0

        # Todo
        # I don't like that epsilon>1 (residual) returns signs of random vector
        #      maybe force to run once? or return vec as zeros in that case?
        #      Seems unintended, but easy to do.

        # Power method iterations
        while itr < max_iters and residual > epsilon:
            itr += 1
            # Todo, this code code actually needs double precision for accuracy... forcing.
            vec_new = self._signs_times_v(Rijs, vec).astype(np.float64, copy=False)
            vec_new = vec_new / norm(vec_new)
            residual = norm(vec_new - vec)
            vec = vec_new
            logger.info(
                f"Iteration {itr}, residual {round(residual, 5)} (target {epsilon})"
            )

        # We need only the signs of the eigenvector
        J_sync = np.sign(vec)
        J_sync = np.sign(J_sync[0]) * J_sync  # Stabilize J_sync

        return J_sync

    def _signs_times_v(self, Rijs, vec):
        """
        Multiplication of the J-synchronization matrix by a candidate eigenvector `vec`

        Wrapper for cpu/gpu dispatch.

        :param Rijs: An n-choose-2x3x3 array of estimates of relative rotations
        :param vec: The current candidate eigenvector of length n-choose-2 from the power method.
        :return: New candidate eigenvector.
        """
        # host/gpu dispatch
        if self.__gpu_module:
            new_vec = self._signs_times_v_cupy(Rijs, vec)
        else:
            new_vec = self._signs_times_v_host(Rijs, vec)

        return new_vec.astype(vec.dtype, copy=False)

    def _signs_times_v_host(self, Rijs, vec):
        """
        See `_signs_times_v`.

        CPU implementation.
        """

        new_vec = np.zeros_like(vec)

        _signs_confs = np.array(
            [[1, 1, 1], [-1, 1, -1], [-1, -1, 1], [1, -1, -1]], dtype=int
        )

        c = np.empty((4))
        desc = "Computing signs_times_v"
        if self.J_weighting:
            desc += " with J_weighting"
        for i in trange(self.n_img - 2, desc=desc):
            for j in range(
                i + 1, self.n_img - 1
            ):  # check bound (taken from MATLAB mex)
                ij = self._pairs_to_linear[i, j]
                Rij = Rijs[ij]
                for k in range(j + 1, self.n_img):
                    ik = self._pairs_to_linear[i, k]
                    jk = self._pairs_to_linear[j, k]
                    Rik = Rijs[ik]
                    Rjk = Rijs[jk]

                    # Compute conjugated rotats
                    Rij_J = J_conjugate(Rij)
                    Rik_J = J_conjugate(Rik)
                    Rjk_J = J_conjugate(Rjk)

                    # Compute R muls and norms
                    c[0] = np.sum(((Rij @ Rjk) - Rik) ** 2)
                    c[1] = np.sum(((Rij_J @ Rjk) - Rik) ** 2)
                    c[2] = np.sum(((Rij @ Rjk_J) - Rik) ** 2)
                    c[3] = np.sum(((Rij @ Rjk) - Rik_J) ** 2)

                    # Find best match
                    best_i = np.argmin(c)
                    best_val = c[best_i]

                    # MATLAB: scores_as_entries == 0
                    s_ij_jk = _signs_confs[best_i][0]
                    s_ik_jk = _signs_confs[best_i][1]
                    s_ij_ik = _signs_confs[best_i][2]

                    # Note there was a third J_weighting option (2) in MATLAB,
                    # but it was not exposed at top level.
                    if self.J_weighting:
                        # MATLAB: scores_as_entries == 1
                        # For each triangle side, find the best alternative
                        alt_ij_jk = c[self._ALTS[0][best_i][0]]
                        if c[self._ALTS[1][best_i][0]] < alt_ij_jk:
                            alt_ij_jk = c[self._ALTS[1][best_i][0]]

                        alt_ik_jk = c[self._ALTS[0][best_i][1]]
                        if c[self._ALTS[1][best_i][1]] < alt_ik_jk:
                            alt_ik_jk = c[self._ALTS[1][best_i][1]]

                        alt_ij_ik = c[self._ALTS[0][best_i][2]]
                        if c[self._ALTS[1][best_i][2]] < alt_ij_ik:
                            alt_ij_ik = c[self._ALTS[1][best_i][2]]

                        # Compute scores
                        s_ij_jk *= 1 - np.sqrt(best_val / alt_ij_jk)
                        s_ik_jk *= 1 - np.sqrt(best_val / alt_ik_jk)
                        s_ij_ik *= 1 - np.sqrt(best_val / alt_ij_ik)

                    # Update vector entries
                    new_vec[ij] += s_ij_jk * vec[jk] + s_ij_ik * vec[ik]
                    new_vec[jk] += s_ij_jk * vec[ij] + s_ik_jk * vec[ik]
                    new_vec[ik] += s_ij_ik * vec[ij] + s_ik_jk * vec[jk]

        return new_vec

    def _signs_times_v_cupy(self, Rijs, vec):
        """
        See `_signs_times_v`.

        CPU implementation.
        """
        import cupy as cp

        signs_times_v = self.__gpu_module.get_function("signs_times_v")

        Rijs_dev = cp.array(Rijs, dtype=np.float64)
        vec_dev = cp.array(vec, dtype=np.float64)
        new_vec_dev = cp.zeros((vec.shape[0]), dtype=np.float64)

        # call the kernel
        blkszx = 512
        nblkx = (self.n_img + blkszx - 1) // blkszx
        signs_times_v(
            (nblkx,),
            (blkszx,),
            (self.n_img, Rijs_dev, vec_dev, new_vec_dev, self.J_weighting),
        )

        # dtoh
        new_vec = new_vec_dev.get().astype(vec.dtype, copy=False)

        return new_vec

    @staticmethod
    def __init_cupy_module():
        """
        Private utility method to read in CUDA source and return as
        compiled CUPY module.
        """

        import cupy as cp

        # Read in contents of file
        fp = os.path.join(os.path.dirname(__file__), "commonline_sync3n.cu")
        with open(fp, "r") as fh:
            module_code = fh.read()

        # CUPY compile the CUDA code
        return cp.RawModule(code=module_code)



"""
Utility wrappers for scipy methods.
"""

import scipy
from packaging.version import Version


def cg(*args, **kwargs):
    """
    Supports scipy cg before and after 1.12.0.
    """

    # older (<1.12.0) scipy cg interface uses `tol` instead of `rtol`.
    # `tol` will be removed in scipy 1.14.0.
    if Version(scipy.__version__) < Version("1.12.0"):
        kwargs["tol"] = kwargs.pop("rtol", None)
    return scipy.sparse.linalg.cg(*args, **kwargs)



import logging
from copy import deepcopy

import numpy as np
from numpy.linalg import solve

from aspire.basis import FFBBasis2D
from aspire.covariance import BatchedRotCov2D
from aspire.denoising import Denoiser
from aspire.noise import WhiteNoiseEstimator
from aspire.optimization import fill_struct
from aspire.utils import mat_to_vec
from aspire.volume import Volume, qr_vols_forward

logger = logging.getLogger(__name__)


def src_wiener_coords(
    sim, mean_vol, eig_vols, lambdas=None, noise_var=0, batch_size=512
):
    """
    Calculate coordinates using Wiener filter

    :param sim: A simulation object containing the images whose coordinates we want.
    :param mean_vol: The mean volume of the source as Volume instance.
    :param eig_vols: The eigenvolumes of the source as Volume instance.
    :param lambdas: The eigenvalues in a K-by-K diagonal matrix (default `eye(K)`).
    :param noise_var: The variance of the noise in the images (default 0).
    :param batch_size: The size of the batches in which to compute the coordinates (default 512).
    :return: A K-by-`src.n` array of coordinates corresponding to the Wiener filter coordinates of each image in sim.

    The coordinates are obtained by the formula
        alpha_s = eig_vols^T H_s ( y_s - P_s mean_vol ) ,

    where P_s is the forward image mapping and y_s is the sth image,
        H_s = Sigma * P_s^T ( P_s Sigma P_s^T + noise_var I )^(-1) ,

    and Sigma is the covariance matrix eig_vols * lambdas * eig_vols^T.
    Note that when noise_var is zero, this reduces to the projecting y_s onto the span of P_s eig_vols.

    # TODO: Find a better place for this functionality other than in utils
    """

    if not isinstance(mean_vol, Volume):
        logger.debug(
            "src_wiener_coords mean_vol should be a Volume instance. Attempt correction."
        )
        if len(mean_vol.shape) == 4 and mean_vol.shape[3] != 1:
            msg = (
                f"Cannot naively convert {mean_vol.shape} to Volume instance."
                "Please change calling code."
            )
            logger.error(msg)
            raise RuntimeError(msg)

        mean_vol = Volume(mean_vol)

    if not isinstance(eig_vols, Volume):
        logger.debug(
            "src_wiener_coords eig_vols should be a Volume instance. Correcting for now."
        )
        eig_vols = Volume(eig_vols)

    if not sim.dtype == mean_vol.dtype == eig_vols.dtype:
        logger.warning(
            "Inconsistent types in src_wiener_coords"
            f" sim {sim.dtype},"
            f" mean_vol {mean_vol.dtype},"
            f" eig_vols {eig_vols.dtype}"
        )

    k = eig_vols.n_vols
    if lambdas is None:
        lambdas = np.eye(k)

    coords = np.zeros((k, sim.n))
    covar_noise = noise_var * np.eye(k)

    for i in range(0, sim.n, batch_size):
        ims = sim.images[i : i + batch_size]
        batch_n = ims.shape[0]
        ims -= sim.vol_forward(mean_vol, i, batch_n)

        Qs, Rs = qr_vols_forward(sim, i, batch_n, eig_vols, k)

        Q_vecs = mat_to_vec(Qs)

        # RCOPT
        ims = np.moveaxis(ims.asnumpy(), 0, 2)
        im_vecs = mat_to_vec(ims)

        for j in range(batch_n):
            im_coords = Q_vecs[:, :, j].T @ im_vecs[:, j]
            covar_im = (Rs[:, :, j] @ lambdas @ Rs[:, :, j].T) + covar_noise
            xx = solve(covar_im, im_coords)
            im_coords = lambdas @ Rs[:, :, j].T @ xx
            coords[:, i + j] = im_coords

    return coords


class DenoiserCov2D(Denoiser):
    """
    Define a derived class for denoising 2D images using Cov2D method
    """

    # Default options for cov2d configuration.
    default_opt = {
        "shrinker": "frobenius_norm",
        "verbose": 0,
        "max_iter": 250,
        "iter_callback": [],
        "store_iterates": False,
        "rel_tolerance": 1e-12,
    }

    def __init__(self, src, basis=None, var_noise=None, batch_size=512, covar_opt=None):
        """
        Initialize an object for denoising 2D images using Cov2D method

        :param src: The source object of 2D images with metadata
        :param basis: The basis method to expand 2D images
        :param var_noise: The estimated variance of noise
        :param batch_size: Integer batch size for processing images.
            Defaults to 512.
        :param covar_opt: Optional dictionary of option overides for Cov2D.
            Provided options will supersede defaults in `DenoiserCov2D.default_opt`.
        """

        super().__init__(src)
        self.batch_size = int(batch_size)

        # When var_noise is not specfically over-ridden,
        #   recompute it now. See #496.
        if var_noise is None:
            logger.info("Estimating noise of images using WhiteNoiseEstimator")
            noise_estimator = WhiteNoiseEstimator(src)
            var_noise = noise_estimator.estimate()
            logger.info(f"Estimated Noise Variance: {var_noise}")
        self.var_noise = var_noise

        if basis is None:
            basis = FFBBasis2D((self.src.L, self.src.L), dtype=src.dtype)

        self.basis = basis
        self.cov2d = None
        self.mean_est = None
        self.covar_est = None

        # Create a local copy of the default options.
        default_opt = deepcopy(self.default_opt)
        # Assign the dtype corresponding to this instance.
        default_opt["precision"] = self.dtype
        # Apply any overrides provided by the user.
        self.covar_opt = fill_struct(covar_opt, default_opt)

        # Initialize the rotationally invariant covariance matrix of 2D images
        # A fixed batch_size is used to loop through image stack.
        self.cov2d = BatchedRotCov2D(self.src, self.basis, batch_size=batch_size)

    def build_denoiser(self):
        """
        Build estimated mean and covariance matrix of 2D images.

        This method should be computed once, on first `images` access.
        """

        if self.covar_est is not None:
            return

        logger.info(f"Building mean estimate for {len(self.src)} images.")
        self.mean_est = self.cov2d.get_mean()

        logger.info(f"Building covariance estimates for {len(self.src)} images.")
        self.covar_est = self.cov2d.get_covar(
            noise_var=self.var_noise,
            mean_coef=self.mean_est,
            covar_est_opt=self.covar_opt,
        )

    def _denoise(self, indices):
        """
        Compute denoised 2D images corresponding to `indices`.

        :return: `Image` object containing denoised images.
        """

        # Lazy evaluate estimates on access.
        # `build_denoiser` internally guards to compute once.
        self.build_denoiser()

        # Denoise requested `indices` selection of 2D images.
        imgs_noise = self.src.images[indices]

        coefs_noise = self.basis.evaluate_t(imgs_noise)
        logger.debug(f"Estimating Cov2D coefficients for {imgs_noise.n_images} images.")
        coefs_estim = self.cov2d.get_cwf_coefs(
            coefs_noise,
            self.cov2d.ctf_basis,
            self.cov2d.ctf_idx[indices],
            mean_coef=self.mean_est,
            covar_coef=self.covar_est,
            noise_var=self.var_noise,
        )

        # Convert Fourier-Bessel coefficients back into 2D images
        logger.info("Converting Cov2D coefficients back to 2D images")
        imgs_denoised = self.basis.evaluate(coefs_estim)

        return imgs_denoised
